--------------------------------------------------------------------------------
-- mod p division. Plus-minus algorithm. Control Unit
--------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;
entity control is
port (
  clk, reset,start: in std_logic;
  p_gt_0, d_equ_0, d_equ_1, sign_d, sign_b, sign_v: in std_logic;
  a_mod_4, u_mod_4, a_plus_b, u_plus_v, u_minus_v: in std_logic_vector(1 downto 0);
  first_step, last_step, ce_a, ce_b, ce_u, ce_v, ce_d, ce_p: out std_logic;
  cont_a: out std_logic_vector(2 downto 0);
  cont_u: out std_logic_vector(5 downto 0);
  cont_d, cont_p: out std_logic_vector(1 downto 0);
  ce_pre: out std_logic; --CE de la precomputacin
  done: out std_logic
);
end control;

architecture rtl of control is

signal current_state: std_logic_vector(1 downto 0);

begin

--state transitions
process(clk, reset)
begin
if reset = '1' then current_state <= "00";
elsif clk'event and clk = '1' then 
	case current_state is
	when "00" => if start = '0' then current_state <= current_state + 1; end if;
	when "01" => if start = '1' then current_state <= current_state + 1; end if;
	when "10" => current_state <= current_state + 1; 
	when others => if p_gt_0 = '0' then current_state <= current_state + 1; end if;
	end case;
end if;
end process;

--control signals
process(current_state, p_gt_0, d_equ_0, sign_d, sign_b, sign_v,
a_mod_4, u_mod_4, a_plus_b, u_plus_v, u_minus_v)
begin
case current_state is

when "00" | "01" =>
	first_step <= '0'; last_step <= '1'; 
	ce_a <= '0'; ce_b <= '0'; ce_u <= '0'; ce_v <= '0'; ce_d <= '0'; ce_p <= '0';
	cont_a <= "000"; cont_d <= "00"; cont_p <= "00"; 
	done <= '1';
--	if (sign_b = '1') and (sign_v = '1') then cont_u <= "000011"; 
--	elsif (sign_b = '1') and (sign_v = '0') then cont_u <= "001011"; 
--	elsif (sign_b = '0') and (sign_v = '1') then cont_u <= "001010"; 
--	else cont_u <= "000010";
--end if;
cont_u <= "000000";
ce_pre <= '0'; --CE de la precomputacin

when "10" =>
first_step <= '1'; last_step <= '0'; 
ce_a <= '1'; ce_b <= '1'; ce_u <= '1'; ce_v <= '1'; ce_d <= '1'; ce_p <= '1';
cont_a <= "000"; cont_u <= "000000"; cont_d <= "10"; cont_p <= "00"; 
done <= '0';
ce_pre <= '1';	  --CE de la precomputacin

when others =>

   ce_pre <= '1'; --CE de la precomputacin

	case a_mod_4 is
	when "00" => 
	
	first_step <= '0'; last_step <= '0'; 
	ce_a <= '1'; ce_b <= '0'; ce_u <= '1'; ce_v <= '0'; ce_d <= '1'; 
	cont_a <= "100"; cont_d <= "00";  
	done <= '0';

	--m mod 4 = 3
		case u_mod_4 is
		when "00" => cont_u <= "100000";
		when "01" => cont_u <= "101000";
		when "10" => cont_u <= "100100";
		when others => cont_u <= "111100";
		end case;

	--m mod 4 = 1
		--case u_mod_4 is
		--when "00" => cont_u <= "100000";
		--when "11" => cont_u <= "101000";
		--when "10" => cont_u <= "100100";
		--when others => cont_u <= "111100";
		--end case;

	if (sign_d = '1') or (d_equ_0 = '1') then ce_p <= '1'; cont_p <= "11";
	elsif d_equ_1 = '1' then ce_p <= '1'; cont_p <= "10";
	else ce_p <= '0'; cont_p <= "--"; --GS
	end if;
	
	when "10" =>
	first_step <= '0'; last_step <= '0'; 
	ce_a <= '1'; ce_b <= '0'; ce_u <= '1'; ce_v <= '0'; ce_d <= '1'; 
	cont_a <= "000"; cont_d <= "01";  
	done <= '0';
	
	if u_mod_4(0) = '0' then cont_u <= "000000"; 
	else cont_u <= "001000"; end if;
	
	if (sign_d = '1') or (d_equ_0 = '1') then  ce_p <= '1'; cont_p <= "10"; 
	else ce_p <= '0'; cont_p <= "--"; end if;

	when others =>  
	
	if a_plus_b = "00" then
	
	first_step <= '0'; last_step <= '0'; 
	ce_a <= '1';  ce_u <= '1'; ce_d <= '1'; 
	cont_a <= "110";	cont_p <= "10";  
	done <= '0';	
	
	--m mod 4 = 3
		case u_plus_v is
		when "00" => cont_u <= "100010";
		when "01" => cont_u <= "101010";
		when "10" => cont_u <= "100110";
		when others => cont_u <= "111110";
		end case;

	--m mod 4 = 3
		--case u_plus_v is
		--when "00" => cont_u <= "100010";
		--when "11" => cont_u <= "101010";
		--when "10" => cont_u <= "100110";
		--when others => cont_u <= "111110";
		--end case;


	if sign_d = '0' then 
		ce_b <= '0'; ce_v <= '0'; cont_d <= "01";
		if d_equ_0 = '1' then ce_p <= '1'; else ce_p <= '0'; end if;
	else
		ce_b <= '1'; ce_v <= '1'; cont_d <= "11";	
		ce_p <= '0';--GS para evitar latches
	end if;
	
	
	else
	--a-b mod 4 = 00
	
	first_step <= '0'; last_step <= '0'; 
	ce_a <= '1';  ce_u <= '1'; ce_d <= '1'; 
	cont_a <= "111";	cont_p <= "10";  
	done <= '0';	
	
	--m mod 4 = 3
		case u_minus_v is
		when "00" => cont_u <= "100011";
		when "01" => cont_u <= "101011";
		when "10" => cont_u <= "100111";
		when others => cont_u <= "111111";
		end case;

	--m mod 4 = 3
		--case u_minus_v is
		--when "00" => cont_u <= "100011";
		--when "11" => cont_u <= "101011";
		--when "10" => cont_u <= "100111";
		--when others => cont_u <= "111111";
		--end case;

	if sign_d = '0' then 
		ce_b <= '0'; ce_v <= '0'; cont_d <= "01";
		if d_equ_0 = '1' then ce_p <= '1'; else ce_p <= '0'; end if;
	else
		ce_b <= '1'; ce_v <= '1'; cont_d <= "11";
    	ce_p <= '0';--GS para evitar latches
	end if;

	end if;
	
	end case;
	
	end case;
	
	end process;


end rtl;

