--------------------------------------------------------------------------------
-- mod p division. Plus-minus algorithm. Datapath
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;
entity data_path is
port (
x, y: in std_logic_vector(n downto 0);
clk, reset: in std_logic; 
first_step, last_step, ce_a, ce_b, ce_u, ce_v, ce_d, ce_p: in std_logic;
cont_a: in std_logic_vector(2 downto 0);
--cont_u: in std_logic_vector(5 downto 0);
cont_d, cont_p: in std_logic_vector(1 downto 0);
z: out std_logic_vector(n-1 downto 0);

p_gt_0, d_equ_0, d_equ_1, sign_d, sign_b, sign_v: out std_logic;
a_mod_4, u_mod_4: inout std_logic_vector(1 downto 0);
a_plus_b, u_plus_v, u_minus_v: out std_logic_vector(1 downto 0);
ce_pre: in std_logic
);
end data_path;

architecture circuit of data_path is

component UV_cell
port (
u, v, x: in STD_LOGIC_VECTOR (N downto 0);
cont_u: in std_logic_vector(5 downto 0);
first_step, last_step: in std_logic;
new_u, new_v: out STD_LOGIC_VECTOR (N downto 0);
sum_0_low: out std_logic_vector(3 downto 0)
);
end component;

component AB_cell
port (
a, b, y: in STD_LOGIC_VECTOR (N downto 0);
cont_a: in std_logic_vector(2 downto 0);
first_step, last_step: in std_logic;
new_a, new_b: out STD_LOGIC_VECTOR (N downto 0)
);
end component;

COMPONENT adjust_z
PORT(
		v : IN std_logic_vector(N downto 0);
		B_neg : IN std_logic;
		v_neg : IN std_logic;          
		res : OUT std_logic_vector(N-1 downto 0)
		);
END COMPONENT;

signal a, b, u, v, new_a, new_b, new_u, new_v: std_logic_vector(n downto 0);
signal d, new_d, p, new_p, not_d, p_minus_1, p_minus_2, dec_d: std_logic_vector(logN downto 0);

signal b_mod_4, v_mod_4: std_logic_vector(1 downto 0);

signal boolean_sum_p: std_logic_vector(logN-1 downto 0);
signal boolean_sum_d: std_logic_vector(logN-2 downto 0);

--Precomputation signals
signal pre_u_mod_4, pre_u_plus_v, pre_u_minus_v, pre_a_plus_b:	std_logic_vector(1 downto 0);
signal int1, sum_0: std_logic_vector(3 downto 0);
signal pcont_u: std_logic_vector(5 downto 0);
signal cont_u: std_logic_vector(5 downto 0);

begin

--uv_block: uv_cell port map (u, v, x, cont_u, first_step, last_step, new_u, new_v, z);
uv_block: uv_cell port map (u, v, x, cont_u, first_step, last_step, new_u, new_v, sum_0);

ab_block: ab_cell port map (a, b, y, cont_a, first_step, last_step, new_a, new_b);

last_adjust: adjust_z PORT MAP( -- sign_b <= b(n); sign_v <= v(n);
		v => new_v, B_neg => b(n), v_neg => v(n),	res => Z	);


p_minus_1 <= p-1; p_minus_2 <= p-2;
p_counter: for i in 0 to logN generate
  new_p(i) <= (not(cont_p(1)) and not(cont_p(0)) and number(i))
               or (cont_p(1) and not(cont_p(0)) and p_minus_1(i))
               or (cont_p(1) and cont_p(0) and p_minus_2(i));
end generate;

with cont_d(0) select dec_d <= d-2 when '0', d-1 when others;

d_counter: for i in 0 to logN generate
  new_d(i) <= (not(cont_d(1)) and dec_d(i))
               or (cont_d(1) and cont_d(0) and not(d(i)));
end generate;

boolean_sum_p(0) <= p(0);
or_gates_p: for i in 1 to logN-1 generate
  boolean_sum_p(i) <= boolean_sum_p(i-1) or p(i);
end generate;

boolean_sum_d(0) <= d(1);
or_gates_d: for i in 1 to logN-2 generate
  boolean_sum_d(i) <= boolean_sum_d(i-1) or d(i+1);
end generate;


p_gt_0 <= (not(p(logN))) and boolean_sum_p(logN - 1);
d_equ_0 <= (not(d(logN))) and (not(boolean_sum_d(logN-2))) and (not(d(0)));
d_equ_1 <= (not(d(logN))) and (not(boolean_sum_d(logN-2))) and d(0);
sign_d <= d(logN);
sign_b <= b(n);
sign_v <= v(n);
a_mod_4 <= a(1 downto 0);
b_mod_4 <= b(1 downto 0);
u_mod_4 <= u(1 downto 0);
v_mod_4 <= v(1 downto 0);
a_plus_b <= a_mod_4 + b_mod_4;
u_plus_v <= u_mod_4 + v_mod_4;
u_minus_v <= u_mod_4 - v_mod_4;

process(clk, reset)
begin
  if reset = '1' then 
     a <= (others => '0');
  elsif clk'event and clk = '1' then
    if ce_a = '1' then 
      if first_step = '1' then 
        a <= y; 
      else
		  a <= new_a; 
	   end if;
	 end if;
  end if;
end process;

process(clk, reset)
begin
  if reset = '1' or first_step = '1' then 
--    b <= (others => '0');
    b <= module(n downto 0);
  elsif clk'event and clk = '1' then
    if ce_b = '1' then 
	   b <= new_b; 
	 end if;
  end if;
end process;

process(clk, reset)
begin
  if reset = '1' then 
    u <= (others => '0');
  elsif clk'event and clk = '1' then
    if ce_u = '1' then 
      if first_step = '1' then 
	     u <= x; 
		else
	     u <= new_u; 
	   end if;
    end if;
  end if;
end process;

process(clk, reset)
begin
  if reset = '1' or first_step = '1' then 
    v <= (others => '0');
  elsif clk'event and clk = '1' then
    if ce_v = '1' then 
      v <= new_v; 
	 end if;
  end if;
end process;

process(clk, reset)
begin
  if reset = '1' then 
    p <= (others => '0');
  elsif clk'event and clk = '1' then
    if ce_p = '1' then 
	   p <= new_p; 
	 end if;
  end if;
end process;

process(clk, reset)
begin
	if reset = '1' then 
	  d <= (others => '0');
	elsif clk'event and clk = '1' then
	  if ce_d = '1' then 
	    d <= new_d; 
	  end if;
	end if;
end process;

---------------------------------------
-- Cont_u precomputation
---------------------------------------
process(clk, reset)
begin
	if reset = '1' then 
	  cont_u <= (others => '0');
	elsif clk'event and clk = '1' then
	    if ce_pre = '1' then
	       cont_u <= pcont_u; 
	    end if;
	end if;
end process;

process(sum_0, cont_u, u, new_a, new_b)
begin
	case cont_u(3 downto 2) is
		when "00" => int1 <= sum_0(3 downto 0);
		when "01" => int1 <= sum_0(3 downto 0) + module(4 downto 1);
		when "10" => int1 <= sum_0(3 downto 0) + module(3 downto 0);
		when others => int1 <= sum_0(3 downto 0) - module(3 downto 0);
	end case;

	pre_a_plus_b <= new_a(1 downto 0) + new_b(1 downto 0); 

end process;

process(int1, cont_u)
begin
	if cont_u(5) = '0' then
		pre_u_mod_4 <= int1(2 downto 1);
	else
		pre_u_mod_4 <= int1(3 downto 2);
	end if;
end process;
pre_u_plus_v <= pre_u_mod_4 + u(1 downto 0);
pre_u_minus_v <= pre_u_mod_4 - u(1 downto 0);


precomp_cont_u: process (new_a, pre_u_mod_4, pre_u_plus_v, pre_u_minus_v)
begin
	case new_a(1 downto 0) is
		when "00" => 
		--m mod 4 = 3
			case pre_u_mod_4 is
				when "00" => pcont_u <= "100000";
				when "01" => pcont_u <= "101000";
				when "10" => pcont_u <= "100100";
				when others => pcont_u <= "111100";
			end case;

		--m mod 4 = 1
			--case u_mod_4 is
			--when "00" => cont_u <= "100000";
			--when "11" => cont_u <= "101000";
			--when "10" => cont_u <= "100100";
			--when others => cont_u <= "111100";
			--end case;

		when "10" =>
--			if u_mod_4(0) = '0' then pcont_u <= "000000"; 
			if pre_u_mod_4(0) = '0' then pcont_u <= "000000"; 
			else pcont_u <= "001000"; end if;
		when others =>  
			if pre_a_plus_b = "00" then
		   --m mod 4 = 3
				--case u_plus_v is
				case pre_u_plus_v is
					when "00" => pcont_u <= "100010";
					when "01" => pcont_u <= "101010";
					when "10" => pcont_u <= "100110";
					when others => pcont_u <= "111110";
				end case;

			--m mod 4 = 3
				--case u_plus_v is
				--when "00" => cont_u <= "100010";
				--when "11" => cont_u <= "101010";
				--when "10" => cont_u <= "100110";
				--when others => cont_u <= "111110";
				--end case;
			else
			--a-b mod 4 = 00
			--m mod 4 = 3
				--case u_minus_v is
				case pre_u_minus_v is
					when "00" => pcont_u <= "100011";
					when "01" => pcont_u <= "101011";
					when "10" => pcont_u <= "100111";
					when others => pcont_u <= "111111";
				end case;

			--m mod 4 = 3
				--case u_minus_v is
				--when "00" => cont_u <= "100011";
				--when "11" => cont_u <= "101011";
				--when "10" => cont_u <= "100111";
				--when others => cont_u <= "111111";
				--end case;
 		end if;
	end case;

end process;















end circuit;

