library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;
entity UV_cell is
  port (
    u, v, x: in STD_LOGIC_VECTOR (N downto 0);
    cont_u: in std_logic_vector(5 downto 0);
    first_step, last_step: in std_logic;
    new_u, new_v: out STD_LOGIC_VECTOR (N downto 0);
    sum_0_low: out std_logic_vector(3 downto 0)
  );
  end UV_cell;

  architecture rtl of UV_cell is
  signal sum_2, sum_4, sum: STD_LOGIC_VECTOR (N downto 0);
  signal op11, op21, sum_0: STD_LOGIC_VECTOR (N+1 downto 0);
  signal op12, op22, sum_1: STD_LOGIC_VECTOR (N+2 downto 0);
begin

  first_adder_operands: for i in 0 to n generate
  op11(i) <= u(i) and not(last_step);
  op21(i) <= cont_u(1) and(cont_u(0) xor v(i));
  end generate;
  op11(n+1) <= op11(n);
  op21(n+1) <= op21(n);

  sum_0 <= op11 + op21 + cont_u(0);

  second_adder_operand: for i in 0 to n+1 generate
  op22(i) <= (not(cont_u(2)) and cont_u(3) and module(i)) 
  or (not(cont_u(3)) and cont_u(2) and two_module(i)) 
  or (cont_u(2) and cont_u(3) and not(module(i)));
  end generate;
  op22(n+2) <= op22(n+1);
  op12 <= sum_0(n+1)&sum_0;

  sum_1 <= op12 + op22 + cont_u(4);

  divide_by_two: for i in 0 to n generate
  sum_2(i) <= sum_1(i+1); 
  end generate;

  divide_by_four: for i in 0 to n generate
  sum_4(i) <= sum_1(i+2); 
  end generate;

  with cont_u(5) select sum <= sum_2 when '0', sum_4 when others;

  with first_step select new_u <= sum when '0', x when others;

  with first_step select new_v <= u when '0', (others => '0') when others;

  --z <= sum_1(n-1 downto 0);
  sum_0_low <= sum_0(3 downto 0);

end rtl;

architecture rtl_opt of UV_cell is
	COMPONENT s_r_n_cell
	PORT(
		u, v : IN std_logic_vector(N downto 0);
		srn : IN std_logic_vector(1 downto 0);          
		res : OUT std_logic_vector(N+1 downto 0)
		);
	END COMPONENT;

	COMPONENT s_r_mux_cell
	PORT(
		s : IN std_logic_vector(N+1 downto 0);
		srn : IN std_logic_vector(1 downto 0);          
		res : OUT std_logic_vector(N+2 downto 0)
		);
	END COMPONENT;

  signal sum_2, sum_4, sum: STD_LOGIC_VECTOR (N downto 0);
  signal sum_0: STD_LOGIC_VECTOR (N+1 downto 0);
  signal sum_1: STD_LOGIC_VECTOR (N+2 downto 0);
begin

  first_adder: s_r_n_cell PORT MAP(
  		u => u, v => v, srn => cont_u(1 downto 0) , res => sum_0	);

  second_adder: s_r_mux_cell PORT MAP(
  		s => sum_0, srn => cont_u(3 downto 2),	res => sum_1 );

  with cont_u(5) select 
       sum <= sum_1(N+1 downto 1) when '0', sum_1(N+2 downto 2) when others;

  new_u <= sum;
  new_v <= u;

  sum_0_low <= sum_0(3 downto 0);

end rtl_opt;